package start;

import java.lang.Exception;
import java.lang.Math;

import java.io.FilenameFilter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.SourceDataLine;
import javax.sound.sampled.FloatControl;

import javax.sound.sampled.Clip;

class TapePlayer extends Thread {
	 byte[] data;
	 byte[] iBinData;
	 int iSamples,iSampleRate;	// total sample count.
	 boolean running=false;
	 DataLine.Info info=null;
	 SourceDataLine sourceDataLine=null;
	 Clip sourceClipLoop=null;
	 File audioFile;
	 AudioInputStream audioInputStream;
	 FileInputStream binInputStream;
	 public FileWriter iDebugLog=null;
	 
	 public final static int kPlayModeNone=0;
	 public final static int kPlayModeWav=1;
	 public final static int kPlayModeTone=2;
	 public final static int kPlayModeBin=3;
	 
	 public final static int kFigFirmware=1;
	 public final static int kTapeRam=2;
	 public final static int kExtFlash=3;

	 int playMode=kPlayModeNone;
	 
	 TapePlayer(AudioFormat audioFormat, AudioInputStream aAudioInputStream) {
		audioInputStream=aAudioInputStream;
		info = new DataLine.Info(SourceDataLine.class, audioFormat);
        try {
			sourceDataLine =  (SourceDataLine) AudioSystem.getLine(info);
			sourceDataLine.open(audioFormat);
			playMode=kPlayModeWav;
        } catch (Exception e) {
          e.printStackTrace();
          return;
        }
    }
    
	TapePlayer(AudioFormat audioFormat, byte[] tone) {
		data=tone;
		info = new DataLine.Info(SourceDataLine.class, audioFormat);
        try {
        	setPriority(Thread.MAX_PRIORITY);
			sourceDataLine =  (SourceDataLine) AudioSystem.getLine(info);
			sourceDataLine.open(audioFormat);
			playMode=kPlayModeTone;
        } catch (Exception e) {
          e.printStackTrace();
          return;
        }
    }

	TapePlayer(AudioFormat audioFormat, byte[] audioData, int samples, int sampleShift) {
		iBinData=audioData;
		iSamples=samples;
		iSampleRate=1<<sampleShift;
		info = new DataLine.Info(SourceDataLine.class, audioFormat);
        try {
        	setPriority(Thread.MAX_PRIORITY);
			sourceDataLine =  (SourceDataLine) AudioSystem.getLine(info);
			sourceDataLine.open(audioFormat);
			playMode=kPlayModeBin;
        } catch (Exception e) {
          e.printStackTrace();
          return;
        }
    }


	 TapePlayer(AudioFormat audioFormat, File aBinFile) {
		audioFile=aBinFile;
		info = new DataLine.Info(SourceDataLine.class, audioFormat);
        try {
			System.out.print(">Bin File player!");
        	binInputStream=new FileInputStream(audioFile);
			sourceDataLine =  (SourceDataLine) AudioSystem.getLine(info);
			sourceDataLine.open(audioFormat);
			playMode=kPlayModeBin;
			System.out.println("<");
        } catch (Exception e) {
          e.printStackTrace();
          return;
        }
    }

	public void setAmp(double gain) {
		FloatControl gainControl = null;
		if(sourceDataLine!=null)
			gainControl=(FloatControl) sourceDataLine.getControl(FloatControl.Type.MASTER_GAIN);
		if(sourceClipLoop!=null)
			gainControl=(FloatControl) sourceClipLoop.getControl(FloatControl.Type.MASTER_GAIN);		
    	float dB = (float) (Math.log(gain) / Math.log(10.0) * 20.0);
    	gainControl.setValue(dB);
	}

	public void playFile() {
	 	long totalBytes=0;
        try {
          int bytesRead = 0;
          while (bytesRead != -1 && running) {
            bytesRead = 
                audioInputStream.read(data, 0, data.length);
            if (bytesRead >= 0) {
              sourceDataLine.write(data, 0, bytesRead);
               totalBytes+=bytesRead;
	 		  //System.out.print(":"+totalBytes);
            }
          }
          System.out.println("Done player");
          audioInputStream=null;
        } catch (IOException e) {
          e.printStackTrace();
          return;
        } finally {
          sourceDataLine.drain();
          sourceDataLine.close();
        }
	}


	public int minReload=0;
	public void playTone() {
		int buffSize=sourceDataLine.available();
		int reloadSize=buffSize/2;
		minReload=reloadSize;
		System.out.println("Tone BufferSize="+buffSize+" reloadSize="+reloadSize);
		//sourceClipLoop.loop(Clip.LOOP_CONTINUOUSLY);
		int pos=0,avail=0,bytesWritten=0;
		while (running) {
			while((avail=sourceDataLine.available())<=reloadSize) {
				try{
					sleep(10);
				}
				catch(Exception ex) {
					System.out.println(" Exception:"+ex);
				}
			}
			// space is available now.
			//System.out.print(" "+sourceDataLine.available());
			if(avail>reloadSize)
				avail=reloadSize&~3;	// maximum data size is reload.
			while(avail>=data.length-pos) {
				bytesWritten=sourceDataLine.write(data, pos, data.length-pos);
				pos=0;
				avail-=bytesWritten; // only write as much as we can.
			}
			//if(avail<minReload)
			//	minReload=avail;
			if(avail>0) {
				bytesWritten=sourceDataLine.write(data, pos, avail);
				pos+=bytesWritten;
				if(pos>=data.length)
					pos=0;
			}
		}
		System.out.println("Done Tone");
		data=null;
	}

	public boolean playing=true;
	
	public void playBin() {
		int buffSize=sourceDataLine.available();
		int reloadSize=buffSize/2;
		data=new byte[16384]; // this is our standard buffer, just under 0.1s
		minReload=reloadSize;
		System.out.println("Bin BufferSize="+buffSize+" reloadSize="+reloadSize+" dataLen="+data.length+ " iSamples="+iSamples+" rate="+iSampleRate);
		System.out.print(" Data={");
		//sourceClipLoop.loop(Clip.LOOP_CONTINUOUSLY);
		int pos=0,avail=0,bitstreamPos=0,ix;
		for(ix=0;ix<iBinData.length;ix++) {
			System.out.print(""+(Integer.toHexString(iBinData[ix]&0xff))+", ");
			if((ix&15)==15)
				System.out.println("");
		}
		
		try{
			sleep(1000); // 0.01s
		}
		catch(Exception ex) {
			System.out.println(" Exception:"+ex);
		}
		while (running && bitstreamPos<iSamples) {
			pos=0;
			if(playing) {
				while(pos<data.length && bitstreamPos<iSamples) {
					if((iBinData[bitstreamPos>>3]&(128>>(bitstreamPos&7)))==0) {
						for(ix=0;ix<iSampleRate;ix++) {
							data[pos+2]=data[pos]=(byte)0;
							data[pos+3]=data[pos+1]=(byte)0x80;	// low output
							pos+=4;
						}
					}
					else {
						for(ix=0;ix<iSampleRate;ix++) {
							data[pos+2]=data[pos]=(byte)0xff;
							data[pos+3]=data[pos+1]=(byte)0x7f;	// high output
							pos+=4;
						}
					}
					bitstreamPos++;
				}
			}
			else {
				// 0,4*iSampleRate==low, 8*iSampleRate,12*iSampleRate==high.
				ix=(data[data.length-1]==data[data.length-4*iSampleRate])? 0:(12*iSampleRate);
				ix+=(data[data.length-1]==0x7f)? 0:(8*iSampleRate);
				for(pos=0;pos<data.length;pos+=4) {
					data[pos+2]=data[pos]=(byte)(((ix&(8*iSampleRate))==0)?0:0xff);
					data[pos+3]=data[pos+1]=(byte)(((ix&(8*iSampleRate))==0)?0x80:0x7f);
					ix+=4;
				}				
			}
			//System.out.println(" Buff: "+bitstreamPos+" pos="+pos);
			while(pos<data.length) {	// at the very end...
				data[pos++]=(byte)0;	// all the rest are 0s.
			}
			//System.out.println(" Padded");
			while((avail=sourceDataLine.available())<=reloadSize) {
				try{
					sleep(10); // 0.01s
				}
				catch(Exception ex) {
					System.out.println(" Exception:"+ex);
				}
			}
			//System.out.println(" About to play");
			sourceDataLine.write(data, 0, data.length); // play whole buffer.
		}
		System.out.println("Done Tone");
		data=null;
	}

	 public void run() {
	 	System.out.println("Starting player");
	 	//if(playMode!=kPlayModeTone)
		sourceDataLine.start();
		running=true;
		playing=true;
		if(data==null && playMode!=kPlayModeBin)
	        data = new byte[(playMode==kPlayModeBin)?2048:1024];// 256 samples.
	    switch(playMode) {
	    case kPlayModeWav:
	    	playFile();
	    	break;
	    case kPlayModeTone:
	    	playTone();
	    	break;
	    case kPlayModeBin:
	    	playBin();
	    	break;
	    }
	    playMode=0;
	 }
	 
	 
	 public void tapeStop() {
	 	System.out.println("minReload="+minReload);
	 	running=false;
	 }

	 public boolean pause(boolean doPause) {
	 	boolean wasPlaying=playing;
	 	playing=!doPause;
	 	return wasPlaying;
	 }
 }


