package start;

import java.lang.Exception;
import java.lang.Math;

import java.io.FilenameFilter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.TargetDataLine;
import javax.sound.sampled.FloatControl;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Clip;

class TapeRecorder extends Thread {
	byte[] data;
	byte[] iBinData;
	byte[] iSampleData;
	int iSamples,iSampleRate;	// total sample count.
	boolean running=false;
	DataLine.Info info=null;
	TargetDataLine iTargetDataLine=null;
	Clip sourceClipLoop=null;
	File audioFile;
	AudioInputStream audioInputStream;
	FileInputStream binInputStream;
	public FileWriter iDebugLog=null;

	public final static int kRecordModeNone=0;
	public final static int kRecordModeWav=1;
	public final static int kRecordModeTone=2;
	public final static int kRecordModeBin=3;

	public final static int kTapeRam=2;
	public final static int kExtFlash=3;

	int playMode=kRecordModeNone;
	public int kSampleDuration=20;
	
    TapeRecorder(AudioFormat audioFormat,  int sampleShift, int aPlayMode) {
		iSampleRate=1<<sampleShift;
		playMode=aPlayMode;
		if(aPlayMode==kRecordModeWav)
			iSampleData=new byte[kSampleDuration*(int)(audioFormat.getSampleRate()+0.5)];
		else
			iSampleData=new byte[131072];	// 128Kb max.
		System.out.println("Tape Record rate="+audioFormat.getSampleRate()+" samples per bit="+iSampleRate);
		DataLine.Info info = new DataLine.Info(TargetDataLine.class, 
						audioFormat); // format is an AudioFormat object
		if (!AudioSystem.isLineSupported(info)) {
			System.out.println(" Data Format Not supported");
			return; 
		}
		// Obtain and open the line.
		try {
			iTargetDataLine = (TargetDataLine) AudioSystem.getLine(info);
			iTargetDataLine.open(audioFormat);
		} catch (LineUnavailableException ex) {
			ex.printStackTrace();
			return;
		}
	}
	
	public void setAmp(double gain) {
		FloatControl gainControl = null;
		if(iTargetDataLine!=null)
			gainControl=(FloatControl) iTargetDataLine.getControl(FloatControl.Type.MASTER_GAIN);
		if(sourceClipLoop!=null)
			gainControl=(FloatControl) sourceClipLoop.getControl(FloatControl.Type.MASTER_GAIN);		
    	float dB = (float) (Math.log(gain) / Math.log(10.0) * 20.0);
    	gainControl.setValue(dB);
	}


	public void recordTone() {
	// Assume that the TargetDataLine, line, has already
		int numBytesRead;
		int totalSamples=0;
		int ix;
		int buffSize=iTargetDataLine.getBufferSize() / 5;
		byte[] recBuff = new byte[buffSize];
		System.out.println("Tone BufferSize="+buffSize);
		
		while (running && totalSamples<iSampleData.length) {
			numBytesRead =  iTargetDataLine.read(recBuff, 0, recBuff.length);
			// Save this chunk of data.
			for(ix=0;ix<numBytesRead && totalSamples<iSampleData.length;ix++)
				iSampleData[totalSamples++]=recBuff[ix];
		}
		
		System.out.println("Done Sampling:");
		iTargetDataLine.stop();	// stop recording.
		for(ix=0;ix<totalSamples-4;ix+=4) {
			int sample=(iSampleData[ix]+(iSampleData[ix+1]<<8))&0xffff;
			if((ix&63)==0)
				System.out.print("\n["+ix+"]");
			System.out.print(" "+Integer.toHexString(sample));
		}
	}

	int kRecBitGlitchTime=2;
	int kRecBitOneMaxPeriod=5;
	public void recordBin() {
	// Assume that the TargetDataLine, line, has already
		int numBytesRead;
		int[] buckets= new int[32];
		int ix,lastCrossing=0,samplePos=0,bitsPos=0,samplePosRd=0,bit=0,last1Pos=0;
		int phase=1;
		int buffSize=iTargetDataLine.getBufferSize() / 5;
		byte[] recBuff = new byte[buffSize];
		System.out.println("Bin BufferSize="+buffSize+" Sample bytes="+iSampleData.length);
		
		while (running && bitsPos<iSampleData.length) {

			numBytesRead =  iTargetDataLine.read(recBuff, 0, recBuff.length);
			// Save this chunk of data, converted into 1s and 0s.
			samplePosRd=samplePos;
			for(ix=0;ix<numBytesRead/4 && (bitsPos>>3)<iSampleData.length;ix++) {
				if(samplePosRd+ix-lastCrossing>=kRecBitGlitchTime) {
					if(recBuff[ix*4+1]*phase<0) {	// phase change.
						bit=((samplePosRd+ix-lastCrossing>kRecBitOneMaxPeriod)?0:1);
						if(samplePosRd+ix-lastCrossing<32)
							buckets[samplePosRd+ix-lastCrossing]++;
						iSampleData[bitsPos>>3]=(byte)((iSampleData[bitsPos>>3]<<1)|bit);
						if(bit==1)
							last1Pos=bitsPos;
						lastCrossing=samplePos;
						phase=-phase;
						bitsPos++;
					}
				}
				samplePos++;
			}
		}
		
		System.out.println("Done Bin Sampling:"+samplePos+" Samples\nBuckets:");
		for(ix=0;ix<32;ix++) {
			System.out.print("["+ix+"]="+buckets[ix]);
		}
		System.out.println("");
		iTargetDataLine.stop();	// stop recording.
		double time=0.0;
		for(ix=0;ix<(last1Pos+8)>>3;ix++) {
			int sample=iSampleData[ix]&0xff;
			if((ix&15)==0)
				System.out.print("\n["+Integer.toHexString(ix)+"("+time+"s)]");
			System.out.print(" "+Integer.toHexString(sample));
			time+=32.0/44100.0;
			for(bit=0;bit<8;bit++) {
				if((sample&(1<<bit))!=0)
					time+=4.0/44100.0;
			}
		}
	}


/*
	public void playBin() {
		int buffSize=iTargetDataLine.available();
		int reloadSize=buffSize/2;
		data=new byte[16384]; // this is our standard buffer, just under 0.1s
		minReload=reloadSize;
		System.out.println("Bin BufferSize="+buffSize+" reloadSize="+reloadSize+" dataLen="+data.length+ " iSamples="+iSamples+" rate="+iSampleRate);
		System.out.print(" Data={");
		//sourceClipLoop.loop(Clip.LOOP_CONTINUOUSLY);
		int pos=0,avail=0,bitstreamPos=0,ix;
		for(ix=0;ix<iBinData.length;ix++) {
			System.out.print(""+(Integer.toHexString(iBinData[ix]&0xff))+", ");
			if((ix&15)==15)
				System.out.println("");
		}
		
		try{
			sleep(1000); // 0.01s
		}
		catch(Exception ex) {
			System.out.println(" Exception:"+ex);
		}
		while (running && bitstreamPos<iSamples) {
			pos=0;
			while(pos<data.length && bitstreamPos<iSamples) {
				if((iBinData[bitstreamPos>>3]&(128>>(bitstreamPos&7)))==0) {
					for(ix=0;ix<iSampleRate;ix++) {
						data[pos+2]=data[pos]=(byte)0;
						data[pos+3]=data[pos+1]=(byte)0x80;	// low output
						pos+=4;
					}
				}
				else {
					for(ix=0;ix<iSampleRate;ix++) {
						data[pos+2]=data[pos]=(byte)0xff;
						data[pos+3]=data[pos+1]=(byte)0x7f;	// high output
						pos+=4;
					}
				}
				bitstreamPos++;
			}
			//System.out.println(" Buff: "+bitstreamPos+" pos="+pos);
			while(pos<data.length) {	// at the very end...
				data[pos++]=(byte)0;	// all the rest are 0s.
			}
			//System.out.println(" Padded");
			while((avail=iTargetDataLine.available())<=reloadSize) {
				try{
					sleep(10); // 0.01s
				}
				catch(Exception ex) {
					System.out.println(" Exception:"+ex);
				}
			}
			//System.out.println(" About to play");
			iTargetDataLine.write(data, 0, data.length); // play whole buffer.
		}
		System.out.println("Done Tone");
		data=null;
	}
*/
	 public void run() {
	 	System.out.println("Starting Recorder");
	 	//if(playMode!=kRecordModeTone)
		iTargetDataLine.start();
		running=true;
	    switch(playMode) {
	    case kRecordModeWav:
	    	break;
	    case kRecordModeTone:
	    	recordTone();
	    	break;
	    case kRecordModeBin:
	    	recordBin();
	    	break;
	    }
	    playMode=0;
	 }
	 
	 public void tapeStop() {
	 	System.out.println("Stopping!");
	 	running=false;
	 }
 }


